

__all__ = ['Stash']


class Stash(list):
    """List that is easy to search for matching contents instead of using an index."""

    @staticmethod
    def compare(list_item, key):
        """Compare each list item with the given key to find the right object in the list."""
        return list_item == key

    def __getitem__(self, key):
        if isinstance(key, int):
            try:
                return super(Stash, self).__getitem__(key)
            except (IndexError, Exception):
                pass

        for list_item in self:
            if self.compare(list_item, key):
                return list_item

        raise KeyError('Invalid item key given!')

    def __setitem__(self, key, value):
        if isinstance(key, int):
            try:
                return super(Stash, self).__setitem__(key, value)
            except (IndexError, Exception):
                pass

        for i, list_item in enumerate(self):
            if self.compare(list_item, key):
                return super(Stash, self).__setitem__(i, value)

        raise KeyError('Invalid item key given!')

    def __contains__(self, key):
        if isinstance(key, int):
            return super(Stash, self).__getitem__(key)

        for list_item in self:
            if self.compare(list_item, key):
                return True

        return False

    def pop(self, key, default=None):
        """Find, remove, and return the given key or default value.

        Args:
            key (int/str/object): Key to use to find the value in the list
            default (object)[None]: Default value to return. If None and key is not found raise a KeyError

        Returns:
            value (object): Returns the value found

        Raises:
            error (KeyError): If default is None and the key was not found.
        """
        if isinstance(key, int):
            try:
                return super(Stash, self).pop(key)
            except (IndexError, Exception):
                pass

        for i, list_item in enumerate(self):
            if self.compare(list_item, key):
                return super(Stash, self).pop(i)

        if default is None:
            raise KeyError('Invalid item key given!')
        return default

    def remove(self, key):
        """Find, remove, and return the given key.

        Args:
            key (int/str/object): Key to use to find the value in the list

        Returns:
            value (object): Returns the value found

        Raises:
            error (ValueError): If default is None and the key was not found.
        """
        if isinstance(key, int):
            try:
                return super(Stash, self).pop(key)
            except (IndexError, Exception):
                pass

        for i, list_item in enumerate(self):
            if self.compare(list_item, key):
                return super(Stash, self).pop(i)

        raise ValueError('list.remove(x): x not in list')



import sys
import threading
import traceback
import contextlib
from collections import OrderedDict

from qtpy import QtCore


__all__ = ['get_updater', 'set_updater', 'ThreadUpdater']


MAIN_UPDATER = None


def get_updater():
    """Return the main updater."""
    global MAIN_UPDATER

    if MAIN_UPDATER is None:
        # Create a default updater
        MAIN_UPDATER = ThreadUpdater()

    return MAIN_UPDATER


def set_updater(updater):
    """Set the main updater."""
    global MAIN_UPDATER
    MAIN_UPDATER = updater


class ThreadUpdater(QtCore.QObject):
    """General timer that will call functions on an interval."""

    starting = QtCore.Signal()
    stopping = QtCore.Signal()

    def __init__(self, timeout=1/30, parent=None):
        super().__init__(parent)
        self._timeout = timeout
        self._running = False
        self._tmr = QtCore.QTimer()
        self._tmr.setSingleShot(False)
        self._tmr.timeout.connect(self.update)

        self._call_latest = OrderedDict()
        self._latest_lock = threading.RLock()
        self._call_in_main = OrderedDict()
        self._main_lock = threading.RLock()
        self._always_call = OrderedDict()
        self._always_lock = threading.RLock()

        self.starting.connect(self.start)
        self.stopping.connect(self.stop)

    @contextlib.contextmanager
    def restart_on_change(self, restart=None):
        """Context manager to stop and restart the timer if it is running."""
        if restart is None:
            restart = self.is_running()
        if restart:
            self.stop()
            yield
            self.start()
        else:
            yield

    def get_timeout(self):
        """Return the update timer interval in seconds."""
        return self._timeout

    def set_timeout(self, value):
        """Set the update timer interval in seconds."""
        with self.restart_on_change(self.is_running()):
            self._timeout = value
            self._tmr.setInterval(int(self.get_timeout() * 1000))

    def is_running(self):
        """Return if running."""
        return self._running

    def stop(self, set_state=True):
        """Stop the updater timer."""
        try:
            self._tmr.stop()
        except:
            pass
        if set_state:
            self._running = False

    def start(self):
        """Start the updater timer."""
        self.stop(set_state=False)
        self._running = True
        self._tmr.start()

    def ensure_running(self):
        """If the updater is not running send a safe signal to start it."""
        if not self.is_running():
            self._running = True
            self.starting.emit()

    def register_continuous(self, func, *args, **kwargs):
        """Register a function to be called on every update continuously."""
        with self._always_lock:
            self._always_call[func] = (args, kwargs)
        self.ensure_running()

    def unregister_continuous(self, func):
        """Unregister a function to be called on every update continuously."""
        with self._always_lock:
            try:
                self._always_call.pop(func, None)
            except:
                pass

    def call_latest(self, func, *args, **kwargs):
        """Call the most recent values for this function in the main thread on the next update call."""
        with self._latest_lock:
            self._call_latest[func] = (args, kwargs)
        self.ensure_running()

    def now_call_latest(self, func, *args, **kwargs):
        """Call the latest value in the main thread. If this is the main thread call now."""
        if threading.current_thread() == threading.main_thread():
            func(*args, **kwargs)
        else:
            self.call_latest(func, *args, **kwargs)

    def call_in_main(self, func, *args, **kwargs):
        """Call this function in the main thread on the next update call."""
        with self._main_lock:
            try:
                self._call_in_main[func].append((args, kwargs))
            except (KeyError, IndexError, Exception):
                self._call_in_main[func] = [(args, kwargs)]
        self.ensure_running()

    def now_call_in_main(self, func, *args, **kwargs):
        """Call in the main thread. If this is the main thread call now."""
        if threading.current_thread() == threading.main_thread():
            func(*args, **kwargs)
        else:
            self.call_in_main(func, *args, **kwargs)

    def update(self):
        """Run the stored function calls."""
        with self._always_lock:
            always = self._always_call.copy()
        with self._latest_lock:
            latest, self._call_latest = self._call_latest, OrderedDict()
        with self._main_lock:
            main, self._call_in_main = self._call_in_main, OrderedDict()

        for func, (args, kwargs) in always.items():
            try:
                func(*args, **kwargs)
            except Exception:
                traceback.print_exc()
                print('Error in {}'.format(func.__name__), file=sys.stderr)

        for func, (args, kwargs) in latest.items():
            try:
                func(*args, **kwargs)
            except Exception:
                traceback.print_exc()
                print('Error in {}'.format(func.__name__), file=sys.stderr)

        for func, li in main.items():
            for args, kwargs in li:
                try:
                    func(*args, **kwargs)
                except Exception:
                    traceback.print_exc()
                    print('Error in {}'.format(func.__name__), file=sys.stderr)



"""
    bt_controller.keymapper
    SeaLandAire Technologies
    @author: jengel
"""
import os
import threading
from qtpy import QtCore, QtGui, QtWidgets

from pyjoystick.interface import Joystick
from pyjoystick.qt.widgets import AxisWidget, ButtonWidget, HatWidget, SpinSlider, LED
from pyjoystick.qt.keys_trigger import RESOURCE_DIR


__all__ = ["JOYSTICK_ACTIVE_IMG", "JOYSTICK_INACTIVE_IMG", "JoystickKeyMapper"]


# ========== ICONS ==========
JOYSTICK_ACTIVE_IMG = os.path.join(RESOURCE_DIR, "gamepad_active.png")
JOYSTICK_INACTIVE_IMG = os.path.join(RESOURCE_DIR, "gamepad_inactive.png")
# ========== END ICONS ==========


class JoystickKeyMapper(QtWidgets.QWidget):
    """Help the user map the keys."""

    Joystick = Joystick

    def __init__(self, joystick=None, parent=None, event_mngr=None):
        super().__init__(parent)

        self._event_mngr = None
        self._joystick = None
        self._joysticks = []
        self.btns_per_row = 5

        # Layout
        self.main_layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.main_layout)

        # Joystick name
        self.name_lbl = QtWidgets.QLabel("Joystick Name")
        self.name_lbl.setSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum)
        font = self.name_lbl.font()
        font.setPointSize(12)
#         font.setWeight(QtWidgets.QFont.Bold)
        self.name_lbl.setFont(font)
        self.name_lay = QtWidgets.QHBoxLayout()
        self.name_lay.addWidget(self.name_lbl, alignment=QtCore.Qt.AlignLeft)
        self.main_layout.addLayout(self.name_lay)
        
        # Button layouts
        lbl = QtWidgets.QLabel("Axes (Analog sticks or triggers):")
        lbl.setToolTip("Values are scaled -100.0 to 100.0 for the slider display.\n"
                       "real values are -1.0 to 1.0")
        self.main_layout.addWidget(lbl)
        self.axis_layout = QtWidgets.QHBoxLayout()
        self.main_layout.addLayout(self.axis_layout)

        self.main_layout.addWidget(QtWidgets.QLabel("Buttons:"))
        self.button_layout = QtWidgets.QVBoxLayout()
        self.main_layout.addLayout(self.button_layout)

        self.main_layout.addWidget(QtWidgets.QLabel("Hats (D-pads):"))
        self.hat_layout = QtWidgets.QHBoxLayout()
        self.main_layout.addLayout(self.hat_layout)

        self.main_layout.addWidget(QtWidgets.QLabel("Track Balls:"))
        self.ball_layout = QtWidgets.QHBoxLayout()
        self.main_layout.addLayout(self.ball_layout)

        # Select joystick button
        self.sel_btn = QtWidgets.QPushButton("Select Joystick")
        self.sel_btn.setSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum)
        self.sel_btn.clicked.connect(self.select_joystick)
        self.name_lay.addWidget(self.sel_btn, alignment=QtCore.Qt.AlignLeft)
        self._dialog = None

        # Items
        self._lock = threading.Lock()
        if joystick is not None:
            self.sel_btn.hide()
            self.set_joystick(joystick)

        if event_mngr is not None:
            self.set_event_mngr(event_mngr)
    # end Constructor

    def get_event_mngr(self):
        """Return the event manager."""
        return self._event_mngr

    def set_event_mngr(self, value):
        """Set the event manager."""
        self._event_mngr = value
        try:
            self._event_mngr.add_joystick = self.add_joystick
            self._event_mngr.remove_joystick = self.remove_joystick
            self._event_mngr.handle_key_event = self.handle_key_event
            self._event_mngr.start()
        except:
            pass

    event_mngr = property(get_event_mngr, set_event_mngr)

    def add_joystick(self, joy):
        """Add the joystick that was found."""
        self._joysticks.append(joy)

    def remove_joystick(self, joy):
        """Remove teh joystick that was removed."""
        try:
            self._joysticks.remove(joy)
        except:
            pass

    def handle_key_event(self, key):
        """Handle the given key that an event occurred on."""
        if self.joystick != key.joystick:
            return

        if key.keytype == key.AXIS:
            self.set_axes_values(key)
        elif key.keytype == key.BUTTON:
            self.set_button_values(key)
        elif key.keytype == key.HAT:
            self.set_hat_values(key)
        elif key.keytype == key.BALL:
            self.set_ball_values(key)

    def get_joysticks(self):
        """Return a list of joysticks."""
        return self._joysticks

    def select_joystick(self):
        """Select a gamepad if there is no joystick selected."""
        try: self._dialog.close()
        except AttributeError: pass

        self._dialog = QtWidgets.QDialog()
        layout = QtWidgets.QVBoxLayout()
        self._dialog.setLayout(layout)
        
        # Joysticks
        btngroup = QtWidgets.QButtonGroup()
        btngroup.setExclusive(True)
        for item in self.get_joysticks():
            btn = QtWidgets.QRadioButton(item.get_name())
            layout.addWidget(btn)
            btngroup.addButton(btn)

        # Accept method        
        def select():
            btn = btngroup.checkedButton()
            joystick = None
            if btn is not None:
                name = btn.text()
                for joy in self.get_joysticks():
                    if joy.get_name() == name:
                        joystick = joy
                        break
            self.set_joystick(joystick)
            self._dialog.close()

        # Buttons
        accept = QtWidgets.QPushButton("Select")
        accept.clicked.connect(select)
        cancel = QtWidgets.QPushButton("Cancel")
        cancel.clicked.connect(self._dialog.close)

        # Button layout
        hlay = QtWidgets.QHBoxLayout()
        hlay.addWidget(accept, alignment=QtCore.Qt.AlignRight)
        hlay.addWidget(cancel, alignment=QtCore.Qt.AlignRight)
        layout.addLayout(hlay)
        self._dialog.show()

    def _clear_layout(self, layout):
        """Clear a layout."""
        while layout.itemAt(0) is not None:
            item = layout.takeAt(0)
            try:
                obj = item.widget()
                if obj is None:
                    obj = item.layout()
                    self._clear_layout(obj)
                    obj.setParent(None)
                    obj.deleteLater()
                else:
                    obj.close()
                    obj.setParent(None)
                    obj.deleteLater()
            except AttributeError:
                pass
            del item

    def removeWidgets(self):
        """Remove all of the testing widgets."""
        self._clear_layout(self.axis_layout)
        self._clear_layout(self.button_layout)
        self._clear_layout(self.hat_layout)
        self._clear_layout(self.ball_layout)

    def createWidgets(self, joystick=None):
        """Initialize the widgets."""
        self.removeWidgets()

        for i in range(joystick.get_numaxes()):
            widg = AxisWidget("Axis "+str(i))
            self.axis_layout.addWidget(widg)

        btn_lay = []
        for i in range(joystick.get_numbuttons()):
            if (i % self.btns_per_row) == 0:
                btn_lay.append(QtWidgets.QHBoxLayout())
                self.button_layout.addLayout(btn_lay[-1])
            widg = ButtonWidget("Button "+str(i)+":")
            btn_lay[-1].addWidget(widg)

        for i in range(joystick.get_numhats()):
            widg = HatWidget("Hat "+str(i)+":")
            self.hat_layout.addWidget(widg)

        for i in range(joystick.get_numballs()):
            widg = HatWidget("Ball "+str(i)+":")
            self.ball_layout.addWidget(widg)

    def get_joystick(self):
        """Return the active joystick."""
        return self._joystick

    def set_joystick(self, joystick):
        """Set the active joystick."""
        if joystick is None:
            self.removeWidgets()
            return

        self._joystick = joystick
        self.createWidgets(self._joystick)
        try:
            self.name_lbl.setText(self._joystick.get_name())
        except:
            pass

    joystick = property(get_joystick, set_joystick)

    def set_axes_values(self, key):
        try:
            self.axis_layout.itemAt(key.number).widget().setValue(key.get_value())
        except AttributeError:
            pass

    def set_button_values(self, key):
        try:
            btn_lay = self.button_layout.itemAt(int(key.number // self.btns_per_row)).layout()
            btn_lay.itemAt(key.number % self.btns_per_row).widget().setValue(key.get_value())
        except AttributeError:
            pass

    def set_hat_values(self, key):
        try:
            self.hat_layout.itemAt(key.number).widget().setValue(key.get_value())
        except AttributeError:
            pass

    def set_ball_values(self, key):
        try:
            self.ball_layout.itemAt(key.number).widget().setValue(key.get_value())
        except AttributeError:
            pass

    def closeEvent(self, *args, **kwargs):
        try:
            self._event_mngr.stop()
        except:
            pass
        return QtWidgets.QWidget.closeEvent(self, *args, **kwargs)


if __name__ == "__main__":
    import sys
    from pyjoystick import ThreadEventManager, ButtonHatRepeater
    from pyjoystick.run_process import MultiprocessingEventManager
    from pyjoystick.sdl2 import Joystick, run_event_loop
    # from pyjoystick.pygame import Joystick, run_event_loop

    app = QtWidgets.QApplication([])

    w = JoystickKeyMapper(event_mngr=ThreadEventManager(run_event_loop, button_repeater=ButtonHatRepeater()))  # , activity_timeout=0.02))
    # w = JoystickKeyMapper(event_mngr=MultiprocessingEventManager(run_event_loop, button_repeater=ButtonHatRepeater()))  # , activity_timeout=0.02))
    w.show()

    sys.exit(app.exec_())



import time
import threading
from pyjoystick.utils import PeriodicThread


__all__ = ['Repeater', 'ButtonRepeater', 'HatRepeater', 'ButtonHatRepeater']


class Repeater(object):
    """Scheduler to Manage multiple button repeats."""

    def __init__(self, first_repeat_timeout=1.0, repeat_timeout=0.5, check_timeout=None, key_repeated=None,
                 get_key_hash=None):
        super().__init__()

        if key_repeated is not None:
            self.key_repeated = key_repeated
        if get_key_hash is not None:
            self.get_key_hash = get_key_hash

        self.first_repeat_timeout = first_repeat_timeout
        self.repeat_timeout = repeat_timeout
        self._check_timeout = check_timeout

        self.thread = None
        self._name = "pyjoystick-ButtonRepeater"
        self._lock = threading.RLock()
        self.key_times = {}

    def key_repeated(self, key):
        """Add the Key to the event queue to be processed."""
        pass

    def get_check_timeout(self):
        """Return the timeout for the thread to check if the keys have repeated. By default this is half of the repeat
        timeout.
        """
        if self._check_timeout is None:
            return self.repeat_timeout / 2
        return self._check_timeout

    def set_check_timeout(self, value):
        """Set the timeout for the thread to check if the keys have repeated. If None is given this will be half of the
        repeat timeout.
        """
        self._check_timeout = value
        try:
            self.thread.interval = self.get_check_timeout()
        except:
            pass

    check_timeout = property(get_check_timeout, set_check_timeout)

    @property
    def name(self):
        """Return the thread name."""
        return self._name

    @name.setter
    def name(self, value):
        """Set the thread name."""
        self._name = value
        try:
            self.thread.name = value
        except:
            pass

    @staticmethod
    def get_key_hash(key):
        """Return the hash for the given key."""
        if key.joystick:
            return '{}:{} {}'.format(key.joystick, key.keytype, key.number)
        else:
            return '{} {}'.format(key.keytype, key.number)

    def start(self):
        """Start the thread to check for button repeats"""
        self.thread = PeriodicThread(self.check_timeout, self._run)
        self.thread.name = self.name
        self.thread.daemon = True
        self.thread.start()

    def stop(self):
        """Stop the thread to check for button repeats"""
        try:
            self.thread.join(0)
        except:
            pass
        self.thread = None

        with self._lock:
            self.key_times = {}

    def set(self, key):
        """Set the key to start or stop repeate based on the value."""
        if key.value:
            self.start_repeat(key)
        else:
            self.stop_repeat(key)

    def start_repeat(self, key):
        """Start a key repeating."""
        with self._lock:
            k = self.get_key_hash(key)
            if k not in self.key_times:
                self.key_times[k] = [time.time() + self.first_repeat_timeout, key]

    def stop_repeat(self, key):
        """Stop a key from repeating."""
        with self._lock:
            try:
                del self.key_times[self.get_key_hash(key)]
            except:
                pass

    def _run(self):
        """Run the event loop continuously."""
        with self._lock:
            for k, (t, key) in self.key_times.items():
                if time.time() > t:
                    new_key = key.copy()
                    new_key.is_repeat = True
                    self.key_repeated(new_key)
                    try:
                        self.key_times[k][0] = time.time() + self.repeat_timeout
                    except:
                        pass


class ButtonRepeater(Repeater):
    def start_repeat(self, key):
        """Start a key repeating."""
        if key.keytype == key.BUTTON:
            Repeater.start_repeat(self, key)


class HatRepeater(Repeater):
    def set(self, key):
        """Set the key to start or stop repeate based on the value."""
        value = key.value

        if key.keytype == key.HAT:
            try:
                k = self.get_key_hash(key)
                with self._lock:
                    if value == 0:
                        # Key needs to stop repeat!
                        self.stop_repeat(key)
                        return
                    elif self.key_times[k][1].value == value:
                        # Hat is repeating already
                        return
                    else:
                        # Hat value changed. Try repeating a different value.
                        self.stop_repeat(key)
                        value = True
            except:
                pass

        if value:
            self.start_repeat(key)
        else:
            self.stop_repeat(key)

    def start_repeat(self, key):
        """Start a key repeating."""
        if key.keytype == key.HAT:
            Repeater.start_repeat(self, key)


class ButtonHatRepeater(HatRepeater):
    def start_repeat(self, key):
        """Start a key repeating."""
        if key.keytype == key.BUTTON or key.keytype == key.HAT:
            Repeater.start_repeat(self, key)



import os
import sys
import time
import threading

from pyjoystick.utils import PYJOYSTICK_DIR, deadband, change_path, rescale

with change_path(PYJOYSTICK_DIR):
    # os.environ.setdefault('SDL_VIDEODRIVER', 'dummy')
    import pygame

from pyjoystick.stash import Stash
from pyjoystick.interface import Key, Joystick as BaseJoystick


__all__ = ['Key', 'Joystick', 'run_event_loop',
           'get_init', 'init', 'quit', 'key_from_event',
           'refresh_joysticks']


PYGAME_LOCK = threading.RLock()


class Joystick(BaseJoystick):
    @classmethod
    def get_joysticks(cls):
        """Return a list of available joysticks."""
        # Check init
        if not get_init():
            init()

        return Stash(cls(i) for i in range(pygame.joystick.get_count()))

    def __new__(cls, identifier=None, *args, **kwargs):
        # Check init
        if not get_init():
            init()

        # Create the object
        joy = super().__new__(cls)

        if identifier is None:
            identifier = 0

        if isinstance(identifier, str):
            for i in range(pygame.joystick.get_count()):
                raw_joystick = pygame.joystick.Joystick(i)
                try:
                    if raw_joystick.get_name() == identifier:
                        joy.joystick = raw_joystick
                        break
                except:
                    pass
        else:
            joy.joystick = pygame.joystick.Joystick(identifier)

        try:
            if not joy.joystick.get_init():
                joy.joystick.init()
        except:
            pass

        try:
            joy.identifier = joy.joystick.get_id()
            joy.name = joy.joystick.get_name()
            joy.numaxes = joy.joystick.get_numaxes()
            joy.numbuttons = joy.joystick.get_numbuttons()
            joy.numhats = joy.joystick.get_numhats()
            joy.numballs = joy.joystick.get_numballs()
            joy.init_keys()
        except:
            pass

        return joy

    def is_active(self):
        """Return if this joystick is still active and available."""
        try:
            return self.joystick.get_init()
        except:
            return False

    def close(self):
        """Close the joystick."""
        try:
            self.joystick.quit()
        except:
            pass


HAT_CONVERTER = {
        (0, 0): Key.HAT_CENTERED,
        (0, 1): Key.HAT_UP,
        (1, 0): Key.HAT_RIGHT,
        (0, -1): Key.HAT_DOWN,
        (-1, 0): Key.HAT_LEFT,
        (1, 1): Key.HAT_RIGHTUP,
        (1, -1): Key.HAT_RIGHTDOWN,
        (-1, 1): Key.HAT_LEFTUP,
        (-1, -1): Key.HAT_LEFTDOWN,
        }


def get_init(module=pygame):
    """Return if the given module is initialize."""
    if module != pygame:
        return module.get_init()
    return pygame.display.get_init()  # Pygame events require the display module


def init(module=pygame):
    """Initialize the given module."""
    if module != pygame:
        module.init()
        return

    # Main pygame init
    if not pygame.display.get_init():
        pygame.display.init()  # Pygame events require the display module
        pygame.joystick.init()

        # if not gui_running():
        pygame.display.set_mode((1, 1), pygame.NOFRAME)  # pygame requires some display.

        # Allowed events
        pygame.event.set_blocked([pygame.QUIT,
                                  pygame.ACTIVEEVENT,
                                  pygame.KEYDOWN,
                                  pygame.KEYUP,
                                  pygame.MOUSEMOTION,
                                  pygame.MOUSEBUTTONUP,
                                  pygame.MOUSEBUTTONDOWN,
                                  pygame.VIDEORESIZE,
                                  pygame.VIDEOEXPOSE,
                                  pygame.USEREVENT
                                  ])

        pygame.event.set_allowed([pygame.JOYAXISMOTION,
                                  pygame.JOYBALLMOTION,
                                  pygame.JOYHATMOTION,
                                  pygame.JOYBUTTONDOWN,
                                  pygame.JOYBUTTONUP])


def quit(module=pygame):
    """Quit and deinitialize the given module."""
    module.quit()


def key_from_event(event, joystick=None):
    """Every library type should implement a key_from_event function to convert an event into a key.

    Args:
        event (pygame.event.Event): Event that occurred
        joystick (Joystick)[None]: Joystick object

    Returns:
        key (Key)[None]: Key created from the event.
    """
    if joystick is None:
        try:
            joystick = Joystick(event.joy)
        except (AttributeError, ValueError):
            return None

    # Check the Joystick
    if event.type == pygame.JOYBUTTONDOWN:
        return Key(Key.BUTTON, event.button, 1, joystick)
    elif event.type == pygame.JOYBUTTONUP:
        return Key(Key.BUTTON, event.button, 0, joystick)
    elif event.type == pygame.JOYAXISMOTION:
        value = event.value
        db = getattr(joystick, 'deadband', 0)
        if db:
            value = deadband(value, db)
        return Key(Key.AXIS, event.axis, value, joystick)
    elif event.type == pygame.JOYHATMOTION:
        value = HAT_CONVERTER.get(tuple(event.value), Key.HAT_CENTERED)
        return Key(Key.HAT, event.hat, value, joystick)
    elif event.type == pygame.JOYBALLMOTION:
        return Key(Key.BALL, event.ball, event.value, joystick)
    return None


def refresh_joysticks(joysticks=None):
    """Refresh the joystick list and return a list of all of the joysticks, new joysticks, and removed joysticks.

    Args:
        joysticks (list/Stash): List of pygame.joystick.Joystick objects.

    Returns:
        sticks (list): List of all pygame.joystick.Joystick objects.
        new (list): List of new joysticks where the names were not in the given list.
        removed (list): List of removed joysticks that were in the given list of joysticks, but were not found.
    """
    if joysticks is None:
        joysticks = []

    sticks = Stash()
    new = {}
    removed = {}

    with PYGAME_LOCK:
        pygame.joystick.quit()
        pygame.joystick.init()

    for i in range(pygame.joystick.get_count()):
        joy = Joystick(i)
        name = joy.get_name()

        if name in joysticks:
            sticks.append(joysticks[name])
        else:
            sticks.append(joy)
            new[name] = joy

    for joy in joysticks:
        name = joy.get_name()
        if name not in sticks:
            removed[name] = joy

    return sticks, new, removed


def get_events():
    try:
        return pygame.event.get(pump=True)
    except:
        pygame.event.pump()
        return pygame.event.get()


def run_event_loop(add_joystick, remove_joystick, handle_key_event, alive=None, refresh_timeout=2, **kwargs):
    """Main function to run and handle events.

    Args:
        add_joystick (callable/function): Called when a new Joystick is found!
        remove_joystick (callable/function): Called when a Joystick is removed!
        handle_key_event (callable/function): Called when a new key event occurs!
        alive (callable/function)[None]: Function to return True to continue running. If None run forever
        refresh_timeout (int)[2]: Timeout for when to refresh the joysticks.
    """
    if alive is None:
        alive = lambda: True

    joysticks = Stash()
    last_refresh = 0  # For disconnect and reconnect

    if not get_init():
        init()

    # pygame.event.set_grab(True)

    while alive():
        try:
            # Refresh timeout
            if (time.time() - last_refresh) > refresh_timeout:
                joysticks, new, removed = refresh_joysticks(joysticks)
                for joy in removed.values():
                    remove_joystick(joy)
                for joy in new.values():
                    add_joystick(joy)
                last_refresh = time.time()

            events = get_events()
            if len(events) > 0:
                # Check events
                for event in events:  # pygame.event.get(pump=True)
                    key = key_from_event(event)
                    if key is not None:
                        handle_key_event(key)
            else:
                time.sleep(0.01)
        except:
            pass
